//
// DCC.C - Interface to the DCC-MB driver
//

#include <stdio.h>
#include <dos.h>

short DCCSetMode(short);
short DCCQueryDriver(short *, short *);
short DCCReset(void);
short DCCActivateLoco(short, short);
short DCCSetSpeed(short, short);
short DCCSetExtraBit(short, short);
short DCCSetDirection(short, short);
short DCCSetNReps(short);
short DCCSetGroup1Acc(short, short, short);
short DCCWritePagedRegister(short, short);
short DCCSetPPortLine(short, short);

dcc_init ()

{
    char    c;
    short   nAcc, nDriverMode, nLocos, nReg, nValue;

    if (DCCQueryDriver(&nDriverMode, &nLocos) != 0xFACE) {
      fprintf(stderr, "\007Error: DCC-MB Driver is not installed.");
      exit (1);
    }

    DCCSetMode(1);
}


//---------------------------------------------------------------------------
//                      DCC-MB Driver Calls
//---------------------------------------------------------------------------

union REGS regs;    // structure that holds values of all registers

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCSetMode(short nMode)
{
    regs.h.ah = 0;
    regs.h.al = (unsigned char)nMode;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCQueryDriver(short *nDriverMode, short *nActiveLocos)
{
    regs.x.ax = 0x00FF;
    int86(0x70, &regs, &regs);
    *nDriverMode = regs.x.bx;
    *nActiveLocos = regs.x.cx;
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCReset()
{
    regs.h.ah = 1;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCActivateLoco(short nDCCAddr, short nSpeedMode)
{
    regs.h.ah = 10;
    regs.h.al = (unsigned char)nSpeedMode;
    regs.x.bx = (unsigned char)nDCCAddr;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCDeactivateLoco(short nDCCAddr)
{
    regs.h.ah = 11;
    regs.x.bx = (unsigned char)nDCCAddr;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCSetSpeed(short nDCCAddr, short nSpeed)
{
    regs.h.ah = 12;
    regs.h.al = (unsigned char)nSpeed;
    regs.h.bl = (unsigned char)nDCCAddr;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCSetDirection(short nDCCAddr, short nDir)
{
    regs.h.ah = 13;
    regs.h.al = (unsigned char)nDir;
    regs.h.bl = (unsigned char)nDCCAddr;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCSetExtraBit(short nDCCAddr, short nValue)
{
    regs.h.ah = 14;
    regs.h.al = (unsigned char)nValue;
    regs.h.bl = (unsigned char)nDCCAddr;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCSetNReps(short nReps)
{
    regs.h.ah = 30;
    regs.h.al = (unsigned char)nReps;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCSetGroup1Acc(short nDCCAddr, short nAccNum, short nValue)
{
    regs.h.ah = 31;
    regs.h.al = (unsigned char)nValue;
    regs.h.bl = (unsigned char)nDCCAddr;
    regs.h.cl = (unsigned char)nAccNum;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCWritePagedRegister(short nValue, short nRegNum)
{
    regs.h.ah = 50;
    regs.h.al = (unsigned char)nValue;
    regs.h.bl = (unsigned char)nRegNum;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
short DCCSetPPortLine(short nValue, short nLineNum)
{
    regs.h.ah = 200;
    regs.h.al = (unsigned char)nValue;
    regs.h.bl = (unsigned char)nLineNum;
    int86(0x70, &regs, &regs);
    return regs.x.ax;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
