


//	Run a train with a remote control.
//	Rs-232 reference from The C Users Journal, April 1991 and
//	December 1991.
//	Uses the DCC-MB driver, which is talked to through the dcc.c module.
//	Copyright 1996, 1997 by Chuck Heller.
//	All rights reserved.

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <string.h>
#include <bios.h>
#include <ctype.h>

#include "keys.h"

#define STR_LEN	40

int	trains;			// Are we talking to trains?
int	keyboard_override;	// Is the keyboard overriding the remote?

char	forward_text [8] = "Forward";
char	reverse_text [8] = "Reverse";

// This is for the infrared rs232 interface.

#define DATA_REG			0x3f8
#define INTERRUPT_ENABLE_REG		0x3f9
#define INTERRUPT_IDENTIFICATION_REG	0x3fa
#define LINE_CONTROL_REG		0x3fb
#define MODEM_CONTROL_REG		0x3fc
#define LINE_STATUS_REG			0x3fd
#define MODEM_STATUS_REG		0x3fe

#define N_TICS		5
		// N_TICS is the number of 55ms tics that we allow to go
		// by before we consider this to be the next key.
		// Increasing N_TICS to 18 would reduce the "autorepeat"
		// rate to once per second, for example.
		// Four would be 55 * 4 or 220ms per key, or about 4 keys
		// per second.

#define IR_BUFFER	255

char ir_buffer [IR_BUFFER];	// ir stands for "Infrared Remote Control"

int  	parity = 0,
	overrun = 0,
	frame = 0,
	break_count = 0,
	lost = 0,
	found = 0,
	next_ir_buffer = 0;// Location of next byte to be written in ir_buffer.

static unsigned long far * tics = (unsigned long far *) 0x0000046c;
								// 55 ms tics

#define KEY_LENGTH	40 + 2
#define N_BUTTONS	30	// CH_UP, CH_DN, VOL_UP, VOL_DN, PAUSE,
				// STOP, etc.

char	keys [N_BUTTONS][KEY_LENGTH + 1] = {0};

int token;	// TOKEN_LOCO_UP, TOKEN_LOCO_DN, etc.

int debug;

int key_found ()

{
	int	i;

	for (i = 0 ; keys[i][0] ; i++)
		if (strstr (ir_buffer, keys[i])) {
			// DEBUG fprintf (stderr, "found key %s\n", keys[i]);
			// DEBUG fprintf (stderr, "in buffer %s\n", ir_buffer);
			return i + 1;
		}
	return 0;	// I didn't find it.
}

long	serial_isr_count = 0;

update_serial_isr_count ()

{
	if (debug) printf ("\033[25H%ld", serial_isr_count);
}

update_token ()

{
	printf ("token %d", token);
}

int get_token ()

//	Wait for the user to press a key on the remote.
//	Then read the token and return it in the global variable token.

{
	int			i;
	static unsigned long	hold_off_until = 0;

	if (keyboard_override)
		return token = get_token_from_keyboard ();

	// DEBUG if (ir_buffer[0])
		// DEBUG printf ("%d, %s\n", next_ir_buffer, ir_buffer);
	ir_buffer [0] = 0;	// Clear the buffer.  It may be full of quite
	next_ir_buffer = 0;	// old "data", since we may have been refreshing
				// the screen for quite some time.
	while (1) {
		if (kbhit ())
			exit_program ();
		// DEBUG update_serial_isr_count ();
		if (*tics >= hold_off_until) {
			if (i = key_found ()) {
				ir_buffer [0] = 0;	// clear the buffer
				next_ir_buffer = 0;
				hold_off_until = *tics + N_TICS;
				return token = i;
			} else if (next_ir_buffer > (IR_BUFFER * 3) / 4) {
				// The ir (infrared) buffer is almost full,
				// but we didn't find any matches.  It's
				// probably full of junk from noise.  Nuke it.

				ir_buffer [0] = 0;	// clear the buffer
				next_ir_buffer = 0;
			}
		} else {
			ir_buffer [0] = 0;	// clear the buffer
			next_ir_buffer = 0;
		}
	}
}

void (interrupt * old_vector)();

void interrupt far serial_isr (void)

{
	char	ch;
	int	temp_index;

	++serial_isr_count;
	_enable ();
	ch = inp (INTERRUPT_IDENTIFICATION_REG);

	switch (ch) {
	    case 6:	// Line status interrupt.
		ch = inp (LINE_STATUS_REG);
	    case 4:	// receive data available
		ch = inp (DATA_REG);
		if (next_ir_buffer < IR_BUFFER - 1) {
			ir_buffer [next_ir_buffer++] = ch ? '1' : '0';
			ir_buffer [next_ir_buffer] = 0;
		}
		break;
	    case 0:	// modem status interrupt
		ch = inp (MODEM_STATUS_REG);
		break;
	}
	outp (0x20, 0x20);	// Tell the interrupt controller we are done
}

install_ir ()

{
	char	ch;

	outp (INTERRUPT_ENABLE_REG, 0xd);
	inp  (DATA_REG);
	inp  (LINE_STATUS_REG);
	old_vector = _dos_getvect (12);
	_dos_setvect (12, serial_isr);

	// enable the interrupt controller to give us the interrupt

	ch = inp (0x21);
	ch &= 0xef;	// IRQ mask for IRQ4 11101111
	outp (0x21, ch);

	// enable the serial chip to give us the interrupts

	outp (MODEM_CONTROL_REG, 0xb);
	if (debug) printf ("install_ir is complete\n");
}

cleanup_ir ()

{
	char	ch;

/*
	outp (INTERRUPT_ENABLE_REG, 0);	// Disable all 8250 interrupts.
	outp (MODEM_CONTROL_REG, 0);	// Turn off more 8250 interrupts.
*/
	// Disable COM1 interrupt
	ch = inp (0x21);
	ch |= ! 0xef;	// IRQ mask for IRQ4 11101111
	outp (0x21, ch);

	_dos_setvect (12, old_vector);
	if (debug) printf ("cleanup_ir is done\n");
}

delay_hack ()

{
/*
	printf ("last %lx, tics %lx, input %d\n",
			last_activity,
			*tics,
			input_index);
*/
}

load_keys ()

/*	Load the database from the file KEYS.TXT.
	This files looks like:

	CH_UP
	00110100101111001011010
	CH_DN
	00110100111111001011000
	VOL_UP
	00110100001111001011110
	VOL_DN
	00110100011111001011100
	PAUSE
	01111001101110000110010
	STOP
	01111000001110000111110
*/

{
	FILE	*fp;
	char	buffer [KEY_LENGTH + 20];
	int	i,
		key;

	fp = fopen ("KEYS.TXT", "r");
	if (!fp) {
		fprintf (stderr, "I could not open the KEYS.TXT file\n");
		exit (1);
	}
	key = 0;
	while (EOF != fscanf (fp, "%60s", buffer)) {
		if (key >= N_BUTTONS) {
			fclose (fp);
			fprintf (stderr, "I found too many entries in the KEYS.TXT file.\nIncrease the value of N_BUTTONS in run.c and give it another try.\n");
			exit (1);
		}
		if (debug) fprintf (stderr, "Key %s\n", buffer);
		if (1 != fscanf (fp, "%60s", buffer)) {
			fclose (fp);
			fprintf (stderr, "I got an EOF in the KEYS.TXT file.\nCheck the file format and give it another try.\n");
			exit (1);
		}
		if (debug) fprintf (stderr, "%.40s\n", buffer);
		strcpy (keys[key], buffer);
		key++;
	}
	fclose (fp);

	// Now load the language file.

	fp = fopen ("WORDS.TXT", "r");
	if (fp) {
		fscanf (fp, "%7s", forward_text);
		fscanf (fp, "%7s", reverse_text);
		fclose (fp);
	}
}

#define MAX_LOCOS	30

int	current_channel = 0;		// The channel# will be a loco number.
					// By loco number, I mean the number
					// painted on the side of the loco,
					// like 1905.
int	current_loco = -1;		// This indexes all the arrays below.
int	n_locos;			// This is the number of locos found
					// in the locos file.

int	DCC_speed [MAX_LOCOS] = {0};
char	DCC_mode [MAX_LOCOS] = {0},	
	DCC_addr [MAX_LOCOS] = {0},
	DCC_dir [MAX_LOCOS] ={1},	// direction, 1 => forward, 0 => reverse
	Loco_number [MAX_LOCOS][5] = {0}, // The number on the side of the loco
	Loco_descrip [MAX_LOCOS][STR_LEN] = {0},
	DCC_accessory [MAX_LOCOS][5] = {0};


#define MAX_BUFFER	100

turn_underscores_into_blanks (char * s)

{
	while (*s) {
		if (*s == '_')
			*s = ' ';
		s++;
	}
}

clear_screen ()

{
	printf ("\033[2J");
}

load_locos ()

//	Load the locomotives database from simple text file.
//	Then activate all these locomotives.
//	When you exit, the global arrays DCC_speed, DCC_mode, DCC_addr, DCC_dir,
//	and Loco_descrip will be loaded, n_locos will be set, current_loco will
//	be 0 (meaning the first loco).
//	The file looks like:

//	DCC Address	DCC Mode	Loco Number	Loco Description
//
//	3		2		1905		NYC 2-6-0 Mogul
//	...

{
	FILE	*fp;
	char	buffer [MAX_BUFFER];
	int	i;

	fp = fopen ("LOCOS.TXT", "r");
	if (!fp) {
		fprintf (stderr, "I could not open the LOCOS.TXT file\n");
		exit (1);
	}
	current_loco = n_locos = 0;
	fgets (buffer, MAX_BUFFER - 1, fp);	// skip the first line
	while (EOF != fscanf (fp, "%d %d %4s", &DCC_addr[n_locos], &DCC_mode[n_locos], &Loco_number [n_locos][0])) {
		fgets (buffer, MAX_BUFFER - 1, fp);
		trim_safe_copy (&Loco_descrip [n_locos][0], STR_LEN, buffer);

		DCC_speed [n_locos] = 0;
		DCC_dir [n_locos] = 1;
		// turn_underscores_into_blanks (&Loco_descrip[n_locos][0]);
		if (++n_locos >= MAX_LOCOS) {
			fclose (fp);
			fprintf (stderr, "I found too many entries in the LOCOS.TXT file.\nIncrease the value of MAX_LOCOS in run.c and give it another try.\n");
			exit (1);
		}
	}
	fclose (fp);
	if (debug) fprintf (stderr, "Addr	Mode	Number	Description\n");
	clear_screen ();
	for (i = 0 ; i < n_locos ; i++) {
		if (trains) DCCActivateLoco (DCC_addr[i], DCC_mode [i]);
		update_loco_line (i);
/*
		if (debug) fprintf (stderr, "%d\t%d\t%s\t%s\n", DCC_addr[i],
						  DCC_mode[i],
						  Loco_number [i],
						  Loco_descrip [i]);
*/
	}
}

max_speed (int loco)

{
	switch (DCC_mode[loco]) {
		case 1:	return 14;
		case 2: return 28;
		case 3: return 128;
	}
}

char	digits [5] = {0};	// The digits of a loco read so far.
				// We will highlight the matches as we go.
int	next_digit = 0;

update_loco_line (int loco)

//	loco is an index into the global arrays Loco_number, etc.
//	Redraw row (loco + 1) on the screen.

{
	int	j;

	if (loco == current_loco)
		printf ("\033[32;7m");	// green, inverse video
	else
		printf ("\033[36m");	// cyan
	printf ("\033[%d;H", loco + 1);	// goto the row
	if (next_digit == 0) 
		// No loco digits are highlighted.
		// Print the whole line, then turn off highlighting,
		// then erase to end.
		printf ("%4.4s %-20s  %7.7s  %d\033[0m\033[K",
			Loco_number [loco],
			Loco_descrip [loco],
			DCC_dir [loco] ? forward_text : reverse_text,
			DCC_speed [loco]);
	else {
		// Some locomotive digits are highlighted.
		// Assume that all we have to paint are the digits.
		// The description, speed, and direction are all OK.

		if (digits_match (loco)) {
			printf ("\033[1m");	// highlighting on
			for (j = 0 ; j < next_digit ; j++)
				printf ("%c", Loco_number[loco][j]);
		} else // redraw the number without highlighting.
			printf ("%4.4s", Loco_number [loco]);
		printf ("\033[0m");	// highlighting and inverse video off
	}
}

redraw_screen ()

{
	int	i;

	for (i = 0 ; i < n_locos ; i++)
		update_loco_line (i);
}

clear_digits ()

{
	if (next_digit) {
		// Some digits are highlighted, so erase the highlighting.
		next_digit = 0;
		redraw_screen ();
	}
	next_digit = 0;
	digits[0] = digits[1] = digits[2] = digits[3] = digits[4] = 0;
}

digits_match (int loco)

{
	int	i;

	for (i = 0 ; digits[i] ; i++)
		if (digits[i] != Loco_number[loco][i])
			return 0;
	return 1;
}

int	previous_loco = 0;	// This is for the "PREV CHANNEL" button

void highlight_digit (int digit)

//	Somebody just pressed a number key, which is in digit.
//	Run through all the locos:
//		if this digit now gives an exact match, set current_loco &leave
//		update the highlighting on the loco numbers
//		if we have no hope of a match, punt

{
	int	loco;
	int	n_matches = 0;
	int	the_match;

	digits [next_digit++] = digit;

	// First look and see if we have a match.

	for (loco = n_locos - 1 ; loco >= 0 ; loco--) {
		if (digits_match (loco)) {
			the_match = loco;
			++n_matches;
		}
	}
	switch (n_matches) {
	    case 1:
		previous_loco = current_loco;
		current_loco = the_match;
		clear_digits ();
		break;
	    case 0:
		clear_digits ();
		break;
	    default:
		if (next_digit >= 4) {
			previous_loco = current_loco;
			current_loco = the_match;
			clear_digits ();
		} else
			for (loco = 0 ; loco < n_locos ; loco++)
				update_loco_line (loco);
		break;
	}
}

parser ()

{
	int		old_loco;

	while (1) {

	get_token ();
	// DEBUG update_token ();
	switch (token) {
	    case TOKEN_LOCO_DN:
		clear_digits ();
		previous_loco = old_loco = current_loco;
		++current_loco;
		if (current_loco >= n_locos)
			current_loco = 0;
		update_loco_line (old_loco);
		update_loco_line (current_loco);
		break;
	    case TOKEN_LOCO_UP:
		clear_digits ();
		previous_loco = old_loco = current_loco;
		--current_loco;
		if (current_loco < 0)
			current_loco = n_locos - 1;
		update_loco_line (old_loco);
		update_loco_line (current_loco);
		break;
	    case TOKEN_SPEED_UP:
		clear_digits ();
		if (DCC_speed [current_loco] < max_speed (current_loco)) {
			++DCC_speed [current_loco];
			update_loco_line (current_loco);
			if (trains) DCCSetSpeed (DCC_addr [current_loco],
				DCC_speed [current_loco]);
		}
		break;
	    case TOKEN_SPEED_DN:
		clear_digits ();
		if (DCC_speed [current_loco] > 0) {
			--DCC_speed [current_loco];
			update_loco_line (current_loco);
			// set speed of loco
			if (trains) DCCSetSpeed (DCC_addr [current_loco],
				DCC_speed [current_loco]);
		}
		break;
	    case TOKEN_0:
	    case TOKEN_1:
	    case TOKEN_2:
	    case TOKEN_3:
	    case TOKEN_4:
	    case TOKEN_5:
	    case TOKEN_6:
	    case TOKEN_7:
	    case TOKEN_8:
	    case TOKEN_9:
		highlight_digit (token - TOKEN_0 + '0');
		break;

	    case TOKEN_PREV_LOCO:
		old_loco = current_loco;
		current_loco = previous_loco;
		previous_loco = old_loco;
		update_loco_line (old_loco);
		update_loco_line (current_loco);
		break;
	    case TOKEN_FORWARD:
		clear_digits ();
		if (DCC_dir [current_loco] == 0) {
			DCC_dir [current_loco] = 1;
			DCC_speed [current_loco] = 0;
			update_loco_line (current_loco);
			if (trains) DCCSetSpeed (DCC_addr [current_loco],
				DCC_speed [current_loco]);
			if (trains) DCCSetDirection (DCC_addr [current_loco],
				DCC_dir [current_loco]);
		}
		break;
	    case TOKEN_REVERSE:
		clear_digits ();
		if (DCC_dir [current_loco] == 1) {
			DCC_dir [current_loco] = 0;
			DCC_speed [current_loco] = 0;
			update_loco_line (current_loco);
			if (trains) DCCSetSpeed (DCC_addr [current_loco],
				DCC_speed [current_loco]);
			if (trains) DCCSetDirection (DCC_addr [current_loco],
				DCC_dir [current_loco]);
		}
		break;
	    case TOKEN_FUNC0:	// Accessory 0
		clear_digits ();
		DCC_accessory[current_loco][0] ^= 1;
		if (trains) DCCSetGroup1Acc (DCC_addr [current_loco], 0,
						DCC_accessory[current_loco][0]);
		break;
	    case TOKEN_FUNC1:	// Accessory 1
		clear_digits ();
		DCC_accessory[current_loco][1] ^= 1;
		if (trains) DCCSetGroup1Acc (DCC_addr [current_loco], 1,
						DCC_accessory[current_loco][1]);
		break;
	    case TOKEN_FUNC2:	// Accessory 2
		clear_digits ();
		DCC_accessory[current_loco][2] ^= 1;
		if (trains) DCCSetGroup1Acc (DCC_addr [current_loco], 2,
						DCC_accessory[current_loco][2]);
		break;
	    case TOKEN_FUNC3:	// Accessory 3
		clear_digits ();
		DCC_accessory[current_loco][3] ^= 1;
		if (trains) DCCSetGroup1Acc (DCC_addr [current_loco], 3,
						DCC_accessory[current_loco][3]);
		break;
	    case TOKEN_FUNC4:	// Accessory 4
		clear_digits ();
		DCC_accessory[current_loco][4] ^= 1;
		if (trains) DCCSetGroup1Acc (DCC_addr [current_loco], 4,
						DCC_accessory[current_loco][4]);
		break;
	    case TOKEN_STOP:
	    case TOKEN_PAUSE:
		clear_digits ();
		if (trains) DCCSetSpeed (DCC_addr [current_loco],
				DCC_speed [current_loco] = 0);
		update_loco_line (current_loco);
		break;
	    case TOKEN_STOP_ALL:
		{
		int	loco;

		clear_digits ();
		if (trains)
			for (loco = 0 ; loco < n_locos ; loco++)
				DCCSetSpeed (DCC_addr [loco], DCC_speed [loco] = 0);
		redraw_screen ();
		}
		break;
	    default:
		fprintf (stderr, "I did not recognize this token: %d\n", token);
		break;
	}

	}
}

exit_program ()

{
	int	i;

	clear_digits ();
	if (trains)
		for (i = 0 ; i < n_locos ; i++)
			DCCSetSpeed (DCC_addr [i], DCC_speed [i] = 0);
	exit (1);
}

main (int argc, char **argv)

{
	long	times,
		i,
		b;
	FILE	*fp;

	fprintf (stderr, "Version 1.0 RUN trains using DCC-MB\n");
	fprintf (stderr, "Copyright Chuck Heller 1996, 1997\n");
	fprintf (stderr, "All rights reserved. chuck_heller@hp.com\n");
	// Print this to standard out so that the user can redirect it.
	printf ("RUN [-t] [-d] [-r]\n");
	printf ("  -t:  do NOT call the DCC-MB driver.\n");
	printf ("  -d:  print many debug messages.\n");
	printf ("  -k:  use the keyboard, not the remote.\n");
	trains = 1;	// Assume that we want to talk to trains.
	debug = 0;	// Assume no debug messages;
	keyboard_override = 0;	// Assume the tokens come from the remote.
	for (i = 1 ; argv[i] ; i++)
		switch (argv[i][1]) {
		    case 'd': debug = 1; break;
		    case 't': trains = 0; break;
		    case 'k': keyboard_override = 1; break;
		    default:  break;
		}
	if (debug)
		printf ("Debug is on -- many messages will print\n");
	if (!trains)
		printf ("The DCC-MB driver will not be called\n");
	if (keyboard_override)
		printf ("The keyboard will be used instead of the remote.\n");
	printf ("Press enter to start running, Ctrl-C to exit.\n");
	getchar ();
	if (!keyboard_override) {
		install_ir ();
		atexit (cleanup_ir);
	}
	load_keys ();
	if (debug) {
		printf ("Press enter to start running, Ctrl-C to exit.\n");
		getchar ();
	}
	if (trains) dcc_init ();
	load_locos ();
	parser ();
}

get_token_from_keyboard ()

{
	unsigned int	key;

	key = _bios_keybrd (_KEYBRD_READ);
	if (key & 0x00ff) {
		// It is just a regular ASCII key.
		// Strip off the scan code in the MSB.
		switch (key &= 0xff) {
		    case 0x1b: exit_program (); break;
		    case 0xd: key = TOKEN_FUNC1; break;
		    case 'F':
		    case 'f': key = TOKEN_FORWARD; break;
		    case 'b':
		    case 'B':
		    case 'R':
		    case 'r': key = TOKEN_REVERSE; break;
		    case 'P':
		    case 'p': key = TOKEN_PREV_LOCO; break;
		    case 'S':
		    case 's': key = TOKEN_STOP; break;
		    case '0': key = TOKEN_0; break;
		    case '1': key = TOKEN_1; break;
		    case '2': key = TOKEN_2; break;
		    case '3': key = TOKEN_3; break;
		    case '4': key = TOKEN_4; break;
		    case '5': key = TOKEN_5; break;
		    case '6': key = TOKEN_6; break;
		    case '7': key = TOKEN_7; break;
		    case '8': key = TOKEN_8; break;
		    case '9': key = TOKEN_9; break;
		    default: fprintf (stderr, ""); break;
		}
	} else { // It is a special key
		switch (key >> 8) {
			case 59: key = TOKEN_FUNC1; break;
			case 60: key = TOKEN_FUNC2; break;
			case 61: key = TOKEN_FUNC3; break;
			case 62: key = TOKEN_FUNC4; break;
			case 72: key = TOKEN_LOCO_UP; break;
			case 80: key = TOKEN_LOCO_DN; break;
			case 77: key = TOKEN_SPEED_UP; break;
			case 75: key = TOKEN_SPEED_DN; break;
			default: key = TOKEN_FUNC0; break;
		}
	}
	return key;
}

trim_safe_copy (char * destin, int length, char * source)

//	The source looks like "     UP Mogul 2-6-2      ".
//	Copy that to the destination without the leading or trailing white
//	space, but don't overflow the destination.

{
	char	*trailer;

	trailer = source + strlen (source) - 1;
	// trailer now points before the null.
	while (trailer > source && isspace (*trailer))
		-- trailer;
	// trailer now points to the last non-blank character.
	// unless of course source is a null string.

	while (isspace(*source))
		++source;
	while ((source <= trailer) && --length)
		*destin++ = *source++;
	*destin = 0;
}
